unit UProducerAsm8051;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, USourceCode, UProducer, Contnrs;

type
  TMemoryType = (mtProgram, mtData, mtEEPROM);

  TDevice = class
    Family: string;
    Memory: array[TMemoryType] of Integer;
  end;

  TAssemblerLine = class
    LabelName: string;
    Instruction: string;
    Operand1: string;
    Operand2: string;
    SourceCode: string;
    function AsString: string;
  end;

  { TProducerAsm8051 }

  TProducerAsm8051 = class(TProducer)
  private
    procedure AddInstruction(LabelName, Instruction, Operand1,
      Operand2: string);
//    procedure GenerateCommonBlock(CommonBlock: TCommonBlock; LabelPrefix: string);
    procedure GenerateExpression(Expression: TExpression; LabelPrefix: string);
    procedure GenerateProgram(ProgramBlock: TProgram);
    procedure GenerateModule(Module: TModule);
  public
    AssemblyCode: TObjectList; // TList<TAssemblerLine>
    procedure AssignToStringList(Target: TStringList); override;
    procedure Produce(Module: TModule); override;
    constructor Create;
    destructor Destroy; override;
  end;


implementation


{ TAssemblerLine }

function TAssemblerLine.AsString: string;
begin
  if LabelName = '' then LabelName := #9 else
    LabelName := LabelName + ':'#9;
  if Operand2 <> '' then Operand1 := Operand1 + ', ';

  Result := LabelName + Instruction + ' ' + Operand1 + Operand2;
end;

{ TProducerAsm8051 }

procedure TProducerAsm8051.AddInstruction(LabelName, Instruction, Operand1,
  Operand2: string);
var
  NewLine: TAssemblerLine;
begin
  NewLine := TAssemblerLine.Create;
  AssemblyCode.Add(NewLine);
  NewLine.LabelName := LabelName;
  NewLine.Instruction := Instruction;
  NewLine.Operand1 := Operand1;
  NewLine.Operand2 := Operand2;
end;


(*
procedure TCommonBlock.GenerateAssembler(Compiler: TCompiler; LabelPrefix: string);
var
  I: Integer;
  LabelName: string;
begin
  with Compiler do
  for I := 0 to Operations.Count - 1 do
  with TOperation(Operations[I]) do begin
    if Referenced then LabelName := Name + '_L' + IntToStr(I)
      else LabelName := '';
    case Instruction of
      inJump: begin
        AddInstruction(LabelName, 'JMP', Name + '_L' + IntToStr(GotoAddress), '');
      end;
      inConditionalJump: begin
        ExpressionTree.GenerateAssembler(Compiler, LabelPrefix + '_L' + IntToStr(GotoAddress));
        AddInstruction(LabelName, 'BRCS', Name + '_L' + IntToStr(GotoAddress), '');
      end;
      inExpressionEvaluation: begin
        if LabelName <> '' then AddInstruction(LabelName, '', '', '');
        ExpressionTree.GenerateAssembler(Compiler, Name + '_L' + IntToStr(GotoAddress));
      end;
      inReturn:
        AddInstruction(LabelName, 'RET', '', '');
    end;
  end;
end;
*)

constructor TProducerAsm8051.Create;
begin
  AssemblyCode := TObjectList.Create;
end;

destructor TProducerAsm8051.Destroy;
begin
  AssemblyCode.Free;
  inherited;
end;

(*
  procedure TProducerAsm8051.GenerateCommonBlock(CommonBlock: TCommonBlock; LabelPrefix: string);
  var
    I: Integer;
    LabelName: string;
  begin
    with CommonBlock do
    for I := 0 to Operations.Count - 1 do
    with TOperation(Operations[I]) do begin
      if Referenced then LabelName := Name + '_L' + IntToStr(I)
        else LabelName := '';
      case Instruction of
        inJump: begin
          AddInstruction(LabelName, 'JMP', Name + '_L' + IntToStr(GotoAddress), '');
        end;
        inConditionalJump: begin
          GenerateExpression(ExpressionTree, LabelPrefix + '_L' + IntToStr(GotoAddress));
          AddInstruction(LabelName, 'BRCS', Name + '_L' + IntToStr(GotoAddress), '');
        end;
        inExpressionEvaluation: begin
          if LabelName <> '' then AddInstruction(LabelName, '', '', '');
          GenerateExpression(ExpressionTree, Name + '_L' + IntToStr(GotoAddress));
        end;
        inReturn:
          AddInstruction(LabelName, 'RET', '', '');
      end;
    end;
  end;
*)

procedure TProducerAsm8051.GenerateExpression(Expression: TExpression; LabelPrefix: string);
var
  I: Integer;
begin
  with Expression do
  case NodeType of
    ntNone: ;
    ntVariable: if Assigned(Variable) then AddInstruction('', 'GETVAR', Variable.Name, '');
    nTFunction: AddInstruction('', 'CALL', FunctionCall.FunctionRef.Name, '');
    ntConstant: AddInstruction('', 'CONST', '', '');
    ntOperator: begin
      for I := 0 to SubItems.Count - 1 do
        GenerateExpression(TExpression(SubItems[I]), LabelPrefix);
      if OperatorName = '+' then AddInstruction('', 'ADD', '', '')
      else if OperatorName = '-' then AddInstruction('', 'SUB', '', '')
      else if OperatorName = '*' then AddInstruction('', 'MUL', '', '')
      else if OperatorName = '/' then AddInstruction('', 'DIV', '', '')
      else if OperatorName = 'div' then AddInstruction('', 'DIV', '', '')
      else if OperatorName = 'mod' then AddInstruction('', 'MOD', '', '')
      else if OperatorName = 'xor' then AddInstruction('', 'XOR', '', '')
      else if OperatorName = 'or' then AddInstruction('', 'OR', '', '')
      else if OperatorName = 'and' then AddInstruction('', 'AND', '', '')
      else if OperatorName = 'not' then AddInstruction('', 'NEG', '', '')
      else if OperatorName = ':=' then AddInstruction('', 'ST', '', '')
      else if OperatorName = '>' then AddInstruction('', 'CP', '', '')
      else if OperatorName = '>=' then AddInstruction('', 'CP', '', '')
      else if OperatorName = '<' then AddInstruction('', 'CP', '', '')
      else if OperatorName = '<=' then AddInstruction('', 'CP', '', '')
      else if OperatorName = '=' then AddInstruction('', 'TST', '', '')
      else if OperatorName = '<>' then AddInstruction('', 'CP', '', '');
    end;
  end;
end;

procedure TProducerAsm8051.GenerateProgram(ProgramBlock: TProgram);
var
  I: Integer;
begin
//  with ProgramBlock do
//  for I := 0 to Modules.Count - 1 do
//    GenerateCommonBlock(TModule(Modules[I]), '');
end;

procedure TProducerAsm8051.GenerateModule(Module: TModule);
begin
  Module.TargetFile := Module.Name + '.asm';
end;

procedure TProducerAsm8051.AssignToStringList(Target: TStringList);
var
  I: Integer;
begin
  Target.Clear;
  for I := 0 to AssemblyCode.Count - 1 do begin
    Target.Add(TAssemblerLine(AssemblyCode[I]).AsString);
  end;
end;

procedure TProducerAsm8051.Produce(Module: TModule);
begin
  GenerateModule(Module);
end;

end.
