unit USourceCode;

{$MODE Delphi}
{$MACRO ON}

interface

uses
  SysUtils, Variants, Classes, Dialogs;

type
  TMemoryType = (mtProgram, mtData, mtEEPROM);

  TModuleType = (mdProgram, mdUnit, mdLibrary, mdPackage);

  TNodeType = (ntNone, ntVariable, ntFunction, ntConstant, ntOperator,
    ntValue, ntTypecast);

  TTypeVisibility = (tvPublic, tvPublished, tvPrivate, tvProtected);

  TValue = Variant; //array of Byte;

  TCommonBlock = class;
  TTypeList = class;
  TConstantList = class;
  TVariableList = class;
  TFunctionList = class;
  TListExpression = class;
  TExpression = class;
  TFunction = class;
  TVariable = class;
  TConstant = class;
  TModule = class;
  TProgram = class;

  TDevice = class
    Family: string;
    Memory: array[TMemoryType] of Integer;
  end;

  TContext = class

  end;

  TCommandList = class;

  TCommand = class
    Parent: TObject;
    CommonBlock: TCommonBlock;
  end;

  TAssignment = class(TCommand)
    Target: TVariable;
    Source: TExpression;
    constructor Create;
    destructor Destroy; override;
  end;

  // TListExpression = TGObjectList<Integer, TExpression>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TExpression}
  {$DEFINE TGObjectListList := TObjectListExpression}
  {$DEFINE TGObjectList := TListExpression}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareExpression}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterExpression}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  { TFunctionCall }

  TFunctionCall = class(TCommand)
    FunctionRef: TFunction;
    ParameterExpression: TListExpression;
    constructor Create;
    destructor Destroy; override;
  end;

  TBeginEnd = class(TCommand)
    Commands: TCommandList;
    CommonBlock: TCommonBlock;
    procedure Clear;
    constructor Create;
    destructor Destroy; override;
  end;

  TWhileDo = class(TCommand)
    Condition: TExpression;
    Command: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TWithDo = class(TCommand)
    Context: TContext;
    Command: TCommand;
  end;

  TRepeatUntil = class(TCommand)
    Block: TCommandList;
    Condition: TExpression;
  end;

  { TForToDo }

  TForToDo = class(TCommand)
    ControlVariable: TVariable;
    Start: TExpression;
    Stop: TExpression;
    Command: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  { TIfThenElse }

  TIfThenElse = class(TCommand)
    Condition: TExpression;
    Command: TCommand;
    ElseCommand: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TCaseOfEndBranche = class
    Constant: TConstant;
    Command: TCommand;
  end;

  // TListCaseOfEndBranche = TGObjectList<Integer, TCaseOfEndBranche>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TCaseOfEndBranche}
  {$DEFINE TGObjectListList := TObjectListCaseOfEndBranche}
  {$DEFINE TGObjectList := TListCaseOfEndBranche}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareCaseOfEndBranche}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterCaseOfEndBranche}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TCaseOfEnd = class(TCommand)
    Expression: TExpression;
    Branches: TListCaseOfEndBranche;
    ElseCommand: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TTryFinally = class(TCommand)
    Block: TCommandList;
    FinallyBlock: TCommandList;
  end;

  TTryExcept = class(TCommand)
    Block: TCommandList;
    ExceptBlock: TCommandList;
  end;

  // TCommandList = TGObjectList<Integer, TCommand>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TCommand}
  {$DEFINE TGObjectListList := TObjectListCommand}
  {$DEFINE TGObjectList := TCommandList}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareCommand}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterCommand}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TCommonBlock = class
    Name: string;
    Parent: TCommonBlock;
    ParentModule: TModule;
    Constants: TConstantList;
    Types: TTypeList;
    Variables: TVariableList;
    Functions: TFunctionList;
    Code: TBeginEnd;
    constructor Create; virtual;
    destructor Destroy; override;
//    procedure CheckReferences;
  end;

  TType = class
    ForwardDeclared: Boolean;
    Internal: Boolean;
    Parent: TTypeList;
    Name: string;
    Size: Integer;
    UsedType: TType;
    Visibility: TTypeVisibility;
  end;

  // TListType = TGObjectList<Integer, TType>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TType}
  {$DEFINE TGObjectListList := TObjectListType}
  {$DEFINE TGObjectList := TListType}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareType}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterType}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TTypeList = class(TListType)
    Parent: TCommonBlock;
    function Search(Name: string; Exported: Boolean = False): TType;
    destructor Destroy; override;
  end;

  TTypeInherited = class(TType)
  end;

  TTypeRecord = class(TType)
    CommonBlock: TCommonBlock;
    constructor Create;
    destructor Destroy; override;
  end;

  TTypeSubRange = class(TType)
    Bottom: Variant;
    Top: Variant;
  end;

  TTypeArray = class(TType)
    IndexType: TType;
    ItemType: TType;
  end;

  TTypePointer = class(TType)
  end;

  TEnumItem = class
    Name: string;
    Index: Integer;
  end;

  // TListEnumItem = TGObjectList<Integer, TEnumItem>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TEnumItem}
  {$DEFINE TGObjectListList := TObjectListEnumItem}
  {$DEFINE TGObjectList := TListEnumItem}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareEnumItem}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterEnumItem}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TTypeEnumeration = class(TType)
    Items: TListEnumItem;
    constructor Create;
    destructor Destroy; override;
  end;

  { TTypeClass }

  TTypeClass = class(TType)
    Items: TListType;
    constructor Create;
    destructor Destroy; override;
  end;

  TConstant = class
    System: Boolean;
    Name: string;
    ValueType: TType;
    Value: TValue;
    Exported: Boolean;
  end;

  // TListConstant = TGObjectList<Integer, TConstant>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TConstant}
  {$DEFINE TGObjectListList := TObjectListConstant}
  {$DEFINE TGObjectList := TListConstant}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareConstant}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterConstant}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TConstantList = class(TListConstant)
    Parent: TCommonBlock;
    function Search(Name: string): TConstant;
    destructor Destroy; override;
  end;

  TVariable = class
    Name: string;
    ValueType: TType;
    DefaultValue: TValue;
    Exported: Boolean;
  end;

  // TListVariable = TGObjectList<Integer, TVariable>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TVariable}
  {$DEFINE TGObjectListList := TObjectListVariable}
  {$DEFINE TGObjectList := TListVariable}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareVariable}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterVariable}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TVariableList = class(TListVariable)
    Parent: TCommonBlock;
    function Search(Name: string; Exported: Boolean = False): TVariable;
    destructor Destroy; override;
  end;

  TParameter = class(TVariable)
  end;

  // TListParameter = TGObjectList<Integer, TParameter>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TParameter}
  {$DEFINE TGObjectListList := TObjectListParameter}
  {$DEFINE TGObjectList := TListParameter}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareParameter}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterParameter}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TParameterList = class(TListParameter)
    Parent: TFunction;
    function Search(Name: string): TParameter;
    destructor Destroy; override;
  end;

  { TExpression }

  TExpression = class
    CommonBlock: TCommonBlock;
    NodeType: TNodeType;
    Variable: TVariable;
    Constant: TConstant;
    UseType: TType;
    FunctionCall: TFunctionCall;
    Value: TValue;
    OperatorName: string;
    SubItems: TListExpression;
    Associated: Boolean;
    Braces: Boolean;
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TExpression);
  end;

  TFunctionType = (ftFunction, ftProcedure, ftConstructor, ftDestructor);

  TFunction = class(TCommonBlock)
  public
    Internal: Boolean;
    FunctionType: TFunctionType;
    Parameters: TParameterList;
    ResultType: TType;
    Exported: Boolean;
    constructor Create; override;
    destructor Destroy; override;
  end;

  // TListFunction = TGObjectList<Integer, TFunction>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TFunction}
  {$DEFINE TGObjectListList := TObjectListFunction}
  {$DEFINE TGObjectList := TListFunction}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareFunction}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterFunction}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TFunctionList = class(TListFunction)
    Parent: TCommonBlock;
    function Search(Name: string; Exported: Boolean = False): TFunction;
    destructor Destroy; override;
  end;

  TUsedModule = class
    Name: string;
    Location: string;
    Module: TModule;
    Exported: Boolean;
  end;

  // TListUsedModule = TGObjectList<Integer, TUsedModule>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TUsedModule}
  {$DEFINE TGObjectListList := TObjectListUsedModule}
  {$DEFINE TGObjectList := TListUsedModule}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareUsedModule}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterUsedModule}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  TUsedModuleList = class(TListUsedModule)
    ParentModule: TModule;
  end;

  { TModule }

  TModule = class
  public
    ParentProgram: TProgram;
    Name: string;
    TargetFile: string;
    UsedModules: TUsedModuleList;
    Body: TCommonBlock;
    function SearchVariable(Name: string; Outside: Boolean): TVariable; virtual;
    function SearchConstant(Name: string; Outside: Boolean): TConstant; virtual;
    function SearchType(AName: string; Outside: Boolean): TType; virtual;
    function SearchFunction(Name: string; Outside: Boolean): TFunction; virtual;
    constructor Create;
    destructor Destroy; override;
  end;

  { TModuleProgram }

  TModuleProgram = class(TModule)
    constructor Create;
    destructor Destroy; override;
  end;

  { TModuleUnit }

  TModuleUnit = class(TModule)
    constructor Create;
    destructor Destroy; override;
  end;

  TModulePackage = class(TModule)

  end;

  TModuleLibrary = class(TModule)
  end;

  // TListModule = TGObjectList<Integer, TModule>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TModule}
  {$DEFINE TGObjectListList := TObjectListModule}
  {$DEFINE TGObjectList := TListModule}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareModule}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterModule}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  { TModuleList }

  TModuleList = class(TListModule)
    function Search(Name: string): TModule;
  end;

  TProgram = class
    Device: TDevice;
    Modules: TModuleList;
    constructor Create;
    destructor Destroy; override;
  end;

const
  KeyWords: array[0..37] of string = ('program', 'unit', 'uses', 'begin', 'end',
    'type', 'const', 'var', 'array', 'record', 'absolute', 'virtual', 'class',
    'set', 'private', 'public', 'interface', 'implementation', 'finalization',
    'initialization', 'for', 'while', 'if', 'case', 'of', 'pointer',
    'object', 'packed', 'procedure', 'function', 'to', 'do', 'downto', 'repeat',
    'until', 'then', 'asm', 'else');
  Operators: array[0..22] of string = ('@', 'not', '*', 'and', '/', 'shl',
    'shr', 'as', 'div', 'mod', 'or', 'xor', '-', '+', '=', '>', '<', '<>', '<=',
     '>=', 'is', 'in', ':=');

implementation

{$DEFINE IMPLEMENTATION_USES}
{$I 'GenericObjectList.inc'}

// TListExpression = TGObjectList<Integer, TExpression>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TExpression}
{$DEFINE TGObjectListList := TObjectListExpression}
{$DEFINE TGObjectList := TListExpression}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareExpression}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterExpression}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListCaseOfEndBranche = TGObjectList<Integer, TCaseOfEndBranche>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TCaseOfEndBranche}
{$DEFINE TGObjectListList := TObjectListCaseOfEndBranche}
{$DEFINE TGObjectList := TListCaseOfEndBranche}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareCaseOfEndBranche}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterCaseOfEndBranche}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TCommandList = TGObjectList<Integer, TCommand>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TCommand}
{$DEFINE TGObjectListList := TObjectListCommand}
{$DEFINE TGObjectList := TCommandList}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareCommand}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterCommand}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListEnumItem = TGObjectList<Integer, TEnumItem>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TEnumItem}
{$DEFINE TGObjectListList := TObjectListEnumItem}
{$DEFINE TGObjectList := TListEnumItem}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareEnumItem}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterEnumItem}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListType = TGObjectList<Integer, TType>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TType}
{$DEFINE TGObjectListList := TObjectListType}
{$DEFINE TGObjectList := TListType}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareType}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterType}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListConstant = TGObjectList<Integer, TConstant>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TConstant}
{$DEFINE TGObjectListList := TObjectListConstant}
{$DEFINE TGObjectList := TListConstant}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareConstant}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterConstant}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListVariable = TGObjectList<Integer, TVariable>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TVariable}
{$DEFINE TGObjectListList := TObjectListVariable}
{$DEFINE TGObjectList := TListVariable}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareVariable}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterVariable}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListParameter = TGObjectList<Integer, TParameter>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TParameter}
{$DEFINE TGObjectListList := TObjectListParameter}
{$DEFINE TGObjectList := TListParameter}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareParameter}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterParameter}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListFunction = TGObjectList<Integer, TFunction>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TFunction}
{$DEFINE TGObjectListList := TObjectListFunction}
{$DEFINE TGObjectList := TListFunction}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareFunction}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterFunction}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListUsedModule = TGObjectList<Integer, TUsedModule>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TUsedModule}
{$DEFINE TGObjectListList := TObjectListUsedModule}
{$DEFINE TGObjectList := TListUsedModule}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareUsedModule}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterUsedModule}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

// TListModule = TGObjectList<Integer, TModule>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TModule}
{$DEFINE TGObjectListList := TObjectListModule}
{$DEFINE TGObjectList := TListModule}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareModule}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterModule}
{$DEFINE IMPLEMENTATION}
{$I 'GenericObjectList.inc'}

{ TFunction }

constructor TFunction.Create;
begin
  inherited;
  Parameters := TParameterList.Create;
  Parameters.Parent := Self;
  //ResultType := TType.Create;
end;

destructor TFunction.Destroy;
begin
  Parameters.Free;
//  ResultType.Free;
  inherited;
end;

{ TProgram }

constructor TProgram.Create;
begin
  Device := TDevice.Create;
  Modules := TModuleList.Create;
end;

destructor TProgram.Destroy;
begin
  Modules.Free;
  Device.Free;
end;

{ TConstant }


{ TConstantList }

destructor TConstantList.Destroy;
begin
  inherited;
end;

function TConstantList.Search(Name: string): TConstant;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(Items[I].Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := Items[I] else begin
    if Assigned(Parent.Parent) then Result := Parent.Parent.Constants.Search(Name)
      else begin
        Result := nil;
      end;
  end;
end;

{ TModule }

constructor TModule.Create;
begin
  inherited;
  UsedModules := TUsedModuleList.Create;
  UsedModules.ParentModule := Self;
  Body := TCommonBlock.Create;
  Body.ParentModule := Self;
end;

destructor TModule.Destroy;
begin
  Body.Free;
  UsedModules.Free;
  inherited;
end;

(*
procedure TCommonBlock.CheckReferences;
var
  I: Integer;
begin
  for I := 0 to Operations.Count - 1 do
  with TOperation(Operations[I]) do begin
    if (Instruction = inJump) or (Instruction = inConditionalJump) then
      TOperation(Operations[GotoAddress]).Referenced := True;
  end;
end;
*)

constructor TCommonBlock.Create;
begin
  Constants := TConstantList.Create;
  Constants.Parent := Self;
  Types := TTypeList.Create;
  Types.Parent := Self;
  Variables := TVariableList.Create;
  Variables.Parent := Self;
  Functions := TFunctionList.Create;
  Functions.Parent := Self;
  Code := TBeginEnd.Create;
  Code.Parent := Self;
  Code.CommonBlock := Self;
end;

destructor TCommonBlock.Destroy;
begin
  Constants.Free;
  Types.Free;
  Variables.Free;
  Functions.Free;
  Code.Free;
  inherited;
end;

{ TTypeList }

destructor TTypeList.Destroy;
begin
  inherited;
end;

function TTypeList.Search(Name: string; Exported: Boolean = False): TType;
var
  I: Integer;
begin
  // Search in own list
  I := 0;
  while (I < Count) and (LowerCase(Items[I].Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := Items[I] else Result := nil;

  // Search in parent
  if not Assigned(Result) and Assigned(Parent.Parent) then
     Result := Parent.Parent.Types.Search(Name);

  // Search global variables
  if not Assigned(Result) and Assigned(Parent.ParentModule) and (not Exported) then begin
    Result := Parent.ParentModule.SearchType(Name, True);
  end;
end;

{ TVariableList }

destructor TVariableList.Destroy;
begin
  inherited;
end;

function TVariableList.Search(Name: string; Exported: Boolean = False): TVariable;
var
  I: Integer;
begin
  // Search in own list
  I := 0;
  while (I < Count) and (LowerCase(Items[I].Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TVariable(Items[I]) else Result := nil;

  if not Exported then begin
    // Search parent block variables
    if not Assigned(Result) then begin
      if Assigned(Parent.Parent) then
        Result := Parent.Parent.Variables.Search(Name)
        else Result := nil;
    end;

    // Search parent function parameters
    if not Assigned(Result) then begin
      if Assigned(Parent) and (Parent is TFunction) then
        Result := TFunction(Parent).Parameters.Search(Name)
        else Result := nil;
    end;

    if not Assigned(Result) then begin
      if Assigned(Parent.Parent) and (Parent.Parent is TFunction) then
        Result := TFunction(Parent.Parent).Parameters.Search(Name)
        else Result := nil;
    end;

    // Search global variables
    //if not Assigned(Result) then begin
    //  if Assigned(Parent.ParentModule) then
    //    Result := Parent.ParentModule.SearchVariable(Name, True);
    //end;
  end;
end;

{ TFunctionList }

destructor TFunctionList.Destroy;
begin
  inherited;
end;

function TFunctionList.Search(Name: string; Exported: Boolean): TFunction;
var
  I: Integer;
begin
  // Search in own list
  I := 0;
  while (I < Count) and (LowerCase(Items[I].Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := Items[I] else Result := nil;

  // Search parent block list
  if not Assigned(Result) and Assigned(Parent.Parent) then
    Result := Parent.Parent.Functions.Search(Name, Exported);

  // Search global variables
  if not Assigned(Result) and Assigned(Parent.ParentModule) and (not Exported) then
    Result := Parent.ParentModule.SearchFunction(Name, True);
end;

{ TExpression }

constructor TExpression.Create;
begin
  SubItems := TListExpression.Create;
  SubItems.Count := 2;
  SubItems.OwnsObjects := False;
end;

destructor TExpression.Destroy;
begin
  SubItems.Free;
  inherited;
end;

procedure TExpression.Assign(Source: TExpression);
begin
  CommonBlock := Source.CommonBlock;
  NodeType := Source.NodeType;
  FunctionCall := Source.FunctionCall;
  Value := Source.Value;
  Associated := Source.Associated;
  OperatorName := Source.OperatorName;
  Variable := Source.Variable;
  SubItems.Assign(Source.SubItems);
end;

{ TParameterList }

destructor TParameterList.Destroy;
begin
  inherited;
end;

function TParameterList.Search(Name: string): TParameter;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(Items[I].Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := Items[I]
    else Result := nil;
end;

{ TBeginEnd }

procedure TBeginEnd.Clear;
begin

end;

constructor TBeginEnd.Create;
begin
  inherited;
  Commands := TCommandList.Create;
end;

destructor TBeginEnd.Destroy;
begin
  Commands.Free;
  inherited;
end;

{ TAssignment }

constructor TAssignment.Create;
begin
//  Source := TExpression.Create;
end;

destructor TAssignment.Destroy;
begin
  Source.Free;
  inherited;
end;

{ TWhileDo }

constructor TWhileDo.Create;
begin
  Condition := TExpression.Create;
  Command := TCommand.Create;
end;

destructor TWhileDo.Destroy;
begin
  Condition.Free;
  Command.Free;
  inherited;
end;

{ TCaseOfEnd }

constructor TCaseOfEnd.Create;
begin
  inherited;
  Branches := TListCaseOfEndBranche.Create
end;

destructor TCaseOfEnd.Destroy;
begin
  Branches.Free;
  inherited;
end;

{ TIfThenElse }

constructor TIfThenElse.Create;
begin
  Condition := TExpression.Create;
end;

destructor TIfThenElse.Destroy;
begin
  Condition.Free;
  inherited Destroy;
end;

{ TFunctionCall }

constructor TFunctionCall.Create;
begin
  inherited;
  ParameterExpression := TListExpression.Create;
end;

destructor TFunctionCall.Destroy;
begin
  ParameterExpression.Free;
  inherited Destroy;
end;

{ TForToDo }

constructor TForToDo.Create;
begin
  inherited;
  Start := TExpression.Create;
  Stop := TExpression.Create;
end;

destructor TForToDo.Destroy;
begin
  Start.Free;
  Stop.Free;;
  inherited Destroy;
end;

{ TTypeRecord }

constructor TTypeRecord.Create;
begin
  inherited;
  CommonBlock := TCommonBlock.Create;
end;

destructor TTypeRecord.Destroy;
begin
  CommonBlock.Free;
  inherited Destroy;
end;

{ TModuleList }

function TModuleList.Search(Name: string): TModule;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(Items[I].Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := Items[I] else Result := nil;
end;

{ TModuleProgram }

function TModule.SearchVariable(Name: string; Outside: Boolean): TVariable;
var
  I: Integer;
begin
  Result := Body.Variables.Search(Name, Outside);

  if not Assigned(Result) and Outside then begin
    I := 0;
    while (I < UsedModules.Count) and (not Assigned(Result)) do begin
      Result := UsedModules[I].Module.SearchVariable(Name, False);
      Inc(I);
    end;
  end;
end;

function TModule.SearchConstant(Name: string; Outside: Boolean): TConstant;
begin

end;

function TModule.SearchType(AName: string; Outside: Boolean): TType;
var
  I: Integer;
begin
  Result := Body.Types.Search(AName, Outside);

  if not Assigned(Result) and Outside then begin
    I := 0;
    while (I < UsedModules.Count) and (not Assigned(Result)) do begin
      with UsedModules[I] do
        if Assigned(Module) then
        with Module do
          Result := SearchType(AName, False);
      Inc(I);
    end;
  end;
end;

function TModule.SearchFunction(Name: string; Outside: Boolean): TFunction;
var
  I: Integer;
begin
  Result := Body.Functions.Search(Name, Outside);

  if not Assigned(Result) and Outside then begin
    I := 0;
    while (I < UsedModules.Count) and (not Assigned(Result)) do begin
    if Assigned(UsedModules[I].Module) then
      Result := UsedModules[I].Module.SearchFunction(Name, False);
      Inc(I);
    end;
  end;
end;

constructor TModuleProgram.Create;
begin
  inherited;
end;

destructor TModuleProgram.Destroy;
begin
  inherited Destroy;
end;

{ TModuleUnit }


constructor TModuleUnit.Create;
begin
  inherited;
end;

destructor TModuleUnit.Destroy;
begin
  inherited Destroy;
end;

{ TTypeEnumeration }

constructor TTypeEnumeration.Create;
begin
  inherited;
  Items := TListEnumItem.Create;
end;

destructor TTypeEnumeration.Destroy;
begin
  Items.Free;
  inherited Destroy;
end;

{ TTypeClass }

constructor TTypeClass.Create;
begin
  Items := TListType.Create;
end;

destructor TTypeClass.Destroy;
begin
  Items.Free;
  inherited Destroy;
end;

end.

